@@ -11,7 +11,14 @@
 
 go blocks are dispatched over an internal thread pool, which
 defaults to 8 threads. The size of this pool can be modified using
-the Java system property `clojure.core.async.pool-size`."
+the Java system property `clojure.core.async.pool-size`.
+
+Set Java system property `clojure.core.async.go-checking` to true
+to validate go blocks do not invoke core.async blocking operations.
+Property is read once, at namespace load time. Recommended for use
+primarily during development. Invalid blocking calls will throw in
+go block threads - use Thread.setDefaultUncaughtExceptionHandler()
+to catch and handle."
   (:refer-clojure :exclude [reduce transduce into merge map take partition
                             partition-by bounded-count])
   (:require [clojure.core.async.impl.protocols :as impl]
@@ -23,9 +30,8 @@ the Java system property `clojure.core.async.pool-size`."
             [clojure.core.async.impl.mutex :as mutex]
             [clojure.core.async.impl.concurrent :as conc]
             )
-  (:import [clojure.core.async ThreadLocalRandom]
-           [java.util.concurrent.locks Lock]
-           [java.util.concurrent Executors Executor]
+  (:import [java.util.concurrent.locks Lock]
+           [java.util.concurrent Executors Executor ThreadLocalRandom]
            [java.util ArrayList]))
 
 (alias 'core 'clojure.core)
@@ -50,6 +56,7 @@ the Java system property `clojure.core.async.pool-size`."
 (defn buffer
   "Returns a fixed buffer of size n. When full, puts will block/park."
   [n]
+  (assert (pos? n) "fixed buffers must have size > 0")
   (buffers/fixed-buffer n))
 
 (defn dropping-buffer
@@ -84,7 +91,6 @@ the Java system property `clojure.core.async.pool-size`."
   ([buf-or-n] (chan buf-or-n nil))
   ([buf-or-n xform] (chan buf-or-n xform nil))
   ([buf-or-n xform ex-handler]
-     (when (and buf-or-n (number? buf-or-n)) (assert (pos? buf-or-n) "fixed buffers must have size > 0"))
      (when xform (assert buf-or-n "buffer must be supplied when transducer is"))
      (channels/chan (if (number? buf-or-n) (buffer buf-or-n) buf-or-n) xform ex-handler)))
 
@@ -93,7 +99,8 @@ the Java system property `clojure.core.async.pool-size`."
   exception-handler. A promise channel can take exactly one value that consumers
   will receive. Once full, puts complete but val is dropped (no transfer).
   Consumers will block until either a value is placed in the channel or the
-  channel is closed. See chan for the semantics of xform and ex-handler."
+  channel is closed, then return the value (or nil) forever. See chan for the
+  semantics of xform and ex-handler."
   ([] (promise-chan nil))
   ([xform] (promise-chan xform nil))
   ([xform ex-handler]
@@ -104,9 +111,23 @@ the Java system property `clojure.core.async.pool-size`."
   [^long msecs]
   (timers/timeout msecs))
 
-(defn <!!
+(defmacro defblockingop
+  [op doc arglist & body]
+  (let [as (mapv #(list 'quote %) arglist)]
+    `(def ~(with-meta op {:arglists `(list ~as) :doc doc})
+       (if (Boolean/getBoolean "clojure.core.async.go-checking")
+         (fn ~arglist
+           (dispatch/check-blocking-in-dispatch)
+           ~@body)
+         (fn ~arglist
+           ~@body)))))
+
+(defblockingop <!!
   "takes a val from port. Will return nil if closed. Will block
-  if nothing is available."
+  if nothing is available.
+  Not intended for use in direct or transitive calls from (go ...) blocks.
+  Use the clojure.core.async.go-checking flag to detect invalid use (see
+  namespace docs)."
   [port]
   (let [p (promise)
         ret (impl/take! port (fn-handler (fn [v] (deliver p v))))]
@@ -124,6 +145,11 @@ the Java system property `clojure.core.async.pool-size`."
   "Asynchronously takes a val from port, passing to fn1. Will pass nil
    if closed. If on-caller? (default true) is true, and value is
    immediately available, will call fn1 on calling thread.
+
+   fn1 may be run in a fixed-size dispatch thread pool and should not
+   perform blocking IO, including core.async blocking ops (those that
+   end in !!).
+
    Returns nil."
   ([port fn1] (take! port fn1 true))
   ([port fn1 on-caller?]
@@ -135,9 +161,12 @@ the Java system property `clojure.core.async.pool-size`."
              (dispatch/run #(fn1 val)))))
        nil)))
 
-(defn >!!
+(defblockingop >!!
   "puts a val into port. nil values are not allowed. Will block if no
-  buffer space is available. Returns true unless port is already closed."
+  buffer space is available. Returns true unless port is already closed.
+  Not intended for use in direct or transitive calls from (go ...) blocks.
+  Use the clojure.core.async.go-checking flag to detect invalid use (see
+  namespace docs)."
   [port val]
   (let [p (promise)
         ret (impl/put! port val (fn-handler (fn [open?] (deliver p open?))))]
@@ -159,8 +188,13 @@ the Java system property `clojure.core.async.pool-size`."
   "Asynchronously puts a val into port, calling fn1 (if supplied) when
    complete, passing false iff port is already closed. nil values are
    not allowed. If on-caller? (default true) is true, and the put is
-   immediately accepted, will call fn1 on calling thread.  Returns
-   true unless port is already closed."
+   immediately accepted, will call fn1 on calling thread.
+
+   fn1 may be run in a fixed-size dispatch thread pool and should not
+   perform blocking IO, including core.async blocking ops (those that
+   end in !!).
+
+   Returns true unless port is already closed."
   ([port val]
      (if-let [ret (impl/put! port val fhnop)]
        @ret
@@ -237,6 +271,7 @@ the Java system property `clojure.core.async.pool-size`."
 (defn do-alts
   "returns derefable [val port] if immediate, nil if enqueued"
   [fret ports opts]
+  (assert (pos? (count ports)) "alts must have at least one channel operation")
   (let [flag (alt-flag)
         n (count ports)
         ^ints idxs (random-array n)
@@ -263,13 +298,15 @@ the Java system property `clojure.core.async.pool-size`."
          (when got
            (channels/box [(:default opts) :default])))))))
 
-(defn alts!!
+(defblockingop alts!!
   "Like alts!, except takes will be made as if by <!!, and puts will
-  be made as if by >!!, will block until completed, and not intended
-  for use in (go ...) blocks."
-  [ports & {:as opts}]
+  be made as if by >!!, will block until completed.
+  Not intended for use in direct or transitive calls from (go ...) blocks.
+  Use the clojure.core.async.go-checking flag to detect invalid use (see
+  namespace docs)."
+  [ports & opts]
   (let [p (promise)
-        ret (do-alts (partial deliver p) ports opts)]
+        ret (do-alts (partial deliver p) ports (apply hash-map opts))]
     (if ret
       @ret
       (deref p))))
@@ -409,6 +446,11 @@ the Java system property `clojure.core.async.pool-size`."
   the only JS thread when in ClojureScript). Upon completion of the
   operation, the body will be resumed.
 
+  go blocks should not (either directly or indirectly) perform operations
+  that may block indefinitely. Doing so risks depleting the fixed pool of
+  go block threads, causing all go block processing to stop. This includes
+  core.async blocking ops (those ending in !!) and other blocking IO.
+
   Returns a channel which will receive the result of the body when
   completed"
   [& body]
@@ -502,14 +544,10 @@ the Java system property `clojure.core.async.pool-size`."
                        true)))]
        (dotimes [_ n]
          (case type
-               :blocking (thread
-                          (let [job (<!! jobs)]
-                            (when (process job)
-                              (recur))))
-               :compute (go-loop []
-                                   (let [job (<! jobs)]
-                                     (when (process job)
-                                       (recur))))
+               (:blocking :compute) (thread
+                                      (let [job (<!! jobs)]
+                                        (when (process job)
+                                          (recur))))
                :async (go-loop []
                                  (let [job (<! jobs)]
                                    (when (async job)
@@ -546,7 +584,7 @@ the Java system property `clojure.core.async.pool-size`."
   should be used for computational parallelism. If you have multiple
   blocking operations to put in flight, use pipeline-blocking instead,
   If you have multiple asynchronous operations to put in flight, use
-  pipeline-async instead."
+  pipeline-async instead. See chan for semantics of ex-handler."
   ([n to xf from] (pipeline n to xf from true))
   ([n to xf from close?] (pipeline n to xf from close? nil))
   ([n to xf from close? ex-handler] (pipeline* n to xf from close? ex-handler :compute)))
@@ -631,14 +669,16 @@ the Java system property `clojure.core.async.pool-size`."
         (recur (inc i) (next s))
         i))))
 
-(defn onto-chan
+(defn onto-chan!
   "Puts the contents of coll into the supplied channel.
 
   By default the channel will be closed after the items are copied,
   but can be determined by the close? parameter.
 
-  Returns a channel which will close after the items are copied."
-  ([ch coll] (onto-chan ch coll true))
+  Returns a channel which will close after the items are copied.
+
+  If accessing coll might block, use onto-chan!! instead"
+  ([ch coll] (onto-chan! ch coll true))
   ([ch coll close?]
      (go-loop [vs (seq coll)]
               (if (and vs (>! ch (first vs)))
@@ -646,14 +686,53 @@ the Java system property `clojure.core.async.pool-size`."
                 (when close?
                   (close! ch))))))
 
-(defn to-chan
+(defn to-chan!
   "Creates and returns a channel which contains the contents of coll,
-  closing when exhausted."
+  closing when exhausted.
+
+  If accessing coll might block, use to-chan!! instead"
   [coll]
   (let [c (bounded-count 100 coll)]
     (if (pos? c)
       (let [ch (chan c)]
-        (onto-chan ch coll)
+        (onto-chan! ch coll)
+        ch)
+      (let [ch (chan)]
+        (close! ch)
+        ch))))
+
+(defn onto-chan
+  "Deprecated - use onto-chan! or onto-chan!!"
+  {:deprecated "1.2"}
+  ([ch coll] (onto-chan! ch coll true))
+  ([ch coll close?] (onto-chan! ch coll close?)))
+
+(defn to-chan
+  "Deprecated - use to-chan! or to-chan!!"
+  {:deprecated "1.2"}
+  [coll]
+  (to-chan! coll))
+
+(defn onto-chan!!
+  "Like onto-chan! for use when accessing coll might block,
+  e.g. a lazy seq of blocking operations"
+  ([ch coll] (onto-chan!! ch coll true))
+  ([ch coll close?]
+   (thread
+     (loop [vs (seq coll)]
+       (if (and vs (>!! ch (first vs)))
+         (recur (next vs))
+         (when close?
+           (close! ch)))))))
+
+(defn to-chan!!
+  "Like to-chan! for use when accessing coll might block,
+  e.g. a lazy seq of blocking operations"
+  [coll]
+  (let [c (bounded-count 100 coll)]
+    (if (pos? c)
+      (let [ch (chan c)]
+        (onto-chan!! ch coll)
         ch)
       (let [ch (chan)]
         (close! ch)
@@ -702,7 +781,6 @@ the Java system property `clojure.core.async.pool-size`."
            (reset! dctr (count chs))
            (doseq [c chs]
              (when-not (put! c val done)
-               (done nil)
                (untap* m c)))
            ;;wait for all
            (when (seq chs)
@@ -758,7 +836,7 @@ the Java system property `clojure.core.async.pool-size`."
         solo-modes #{:mute :pause}
         attrs (conj solo-modes :solo)
         solo-mode (atom :mute)
-        change (chan)
+        change (chan (sliding-buffer 1))
         changed #(put! change true)
         pick (fn [attr chs]
                (reduce-kv
@@ -999,7 +1077,7 @@ the Java system property `clojure.core.async.pool-size`."
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; deprecated - do not use ;;;;;;;;;;;;;;;;;;;;;;;;;
 (defn map<
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   [f ch]
   (reify
    impl/Channel
@@ -1031,7 +1109,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn map>
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   [f ch]
   (reify
    impl/Channel
@@ -1047,7 +1125,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn filter>
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   [p ch]
   (reify
    impl/Channel
@@ -1065,13 +1143,13 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn remove>
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   [p ch]
   (filter> (complement p) ch))
 
 (defn filter<
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([p ch] (filter< p ch nil))
   ([p ch buf-or-n]
      (let [out (chan buf-or-n)]
@@ -1086,7 +1164,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn remove<
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([p ch] (remove< p ch nil))
   ([p ch buf-or-n] (filter< (complement p) ch buf-or-n)))
 
@@ -1102,7 +1180,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn mapcat<
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([f in] (mapcat< f in nil))
   ([f in buf-or-n]
     (let [out (chan buf-or-n)]
@@ -1111,7 +1189,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn mapcat>
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([f out] (mapcat> f out nil))
   ([f out buf-or-n]
      (let [in (chan buf-or-n)]
@@ -1120,7 +1198,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn unique
  "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([ch]
      (unique ch nil))
   ([ch buf-or-n]
@@ -1138,7 +1216,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn partition
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([n ch]
      (partition n ch nil))
   ([n ch buf-or-n]
@@ -1163,7 +1241,7 @@ the Java system property `clojure.core.async.pool-size`."
 
 (defn partition-by
   "Deprecated - this function will be removed. Use transducer instead"
-  {:deprecated "0.1.319.0-6b1aca-alpha"}
+  {:deprecated "0.1.319.0-6b1aca-alpha", :skip-wiki true}
   ([f ch]
      (partition-by f ch nil))
   ([f ch buf-or-n]
